@@ -550,7 +550,7 @@ BOOL X11DRV_ProcessEvents( DWORD mask )
     XFlush( gdi_display );
     if (count) TRACE( "processed %d events\n", count );
 
-    if (check_fd_events( ConnectionNumber( data->display ), POLLIN )) return FALSE;
+    if (mask != QS_ALLINPUT || check_fd_events( ConnectionNumber( data->display ), POLLIN )) return FALSE;
     XFlush( data->display ); /* all events have been processed, flush any pending request */
     return TRUE;
 }
@@ -1352,12 +1352,25 @@ static BOOL X11DRV_PropertyNotify( HWND hwnd, XEvent *xev )
 void X11DRV_ActivateWindow( HWND hwnd, HWND previous )
 {
     struct x11drv_win_data *data;
+    BOOL flush = FALSE;
 
-    if (!is_virtual_desktop()) set_net_active_window( hwnd, previous );
+    if (!is_virtual_desktop())
+    {
+        set_net_active_window( hwnd, previous );
+        flush = TRUE;
+    }
 
-    if (!(data = get_win_data( hwnd ))) return;
-    if (!data->managed || data->embedder) set_input_focus( data );
-    release_win_data( data );
+    if ((data = get_win_data( hwnd )))
+    {
+        if (!data->managed || data->embedder)
+        {
+            set_input_focus( data );
+            flush = TRUE;
+        }
+        release_win_data( data );
+    }
+
+    if (flush) XFlush( x11drv_thread_data()->display );
 }
 
 static void drag_drop_enter( UINT entries_size, struct format_entry *entries )
