@@ -38,6 +38,9 @@
 #include <X11/extensions/XInput2.h>
 #endif
 
+#undef Status  /* avoid conflict with wintrnl.h */
+typedef int Status;
+
 #define BOOL X_BOOL
 #define BYTE X_BYTE
 #define INT8 X_INT8
@@ -46,6 +49,12 @@
 #define INT64 X_INT64
 #include <X11/Xmd.h>
 #include <X11/Xproto.h>
+#ifdef HAVE_X11_EXTENSIONS_XF86VMODE_H
+#include <X11/extensions/xf86vmode.h>
+#endif
+#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
+#include <X11/extensions/Xrandr.h>
+#endif
 #undef BOOL
 #undef BYTE
 #undef INT8
@@ -54,17 +63,15 @@
 #undef INT64
 #undef LONG64
 
-#undef Status  /* avoid conflict with wintrnl.h */
-typedef int Status;
-
 /* avoid conflict with processthreadsapi.h */
 #undef ControlMask
 
 #include "windef.h"
 #include "winbase.h"
 #include "ntgdi.h"
+#include "shlobj.h"
+#include "wine/unixlib.h"
 #include "wine/gdi_driver.h"
-#include "unixlib.h"
 #include "wine/list.h"
 #include "wine/debug.h"
 #include "mwm.h"
@@ -247,10 +254,11 @@ extern BOOL X11DRV_GetWindowStateUpdates( HWND hwnd, UINT *state_cmd, UINT *swp_
 extern BOOL X11DRV_CreateWindowSurface( HWND hwnd, BOOL layered, const RECT *surface_rect, struct window_surface **surface );
 extern void X11DRV_MoveWindowBits( HWND hwnd, const struct window_rects *old_rects,
                                    const struct window_rects *new_rects, const RECT *valid_rects );
-extern void X11DRV_WindowPosChanged( HWND hwnd, HWND insert_after, HWND owner_hint, UINT swp_flags, BOOL fullscreen,
+extern void X11DRV_WindowPosChanged( HWND hwnd, HWND insert_after, HWND owner_hint, UINT swp_flags,
                                      const struct window_rects *new_rects, struct window_surface *surface );
 extern BOOL X11DRV_SystemParametersInfo( UINT action, UINT int_param, void *ptr_param,
                                          UINT flags );
+extern LRESULT X11DRV_WintabProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam, void *buffer );
 extern void X11DRV_ThreadDetach(void);
 
 /* X11 driver internal functions */
@@ -665,6 +673,7 @@ struct x11drv_win_data
     UINT        is_offscreen : 1; /* has been moved offscreen by the window manager */
     UINT        parent_invalid : 1; /* is the parent host window possibly invalid */
     UINT        reparenting : 1; /* window is being reparented, likely from a decoration change */
+    UINT        is_resizable : 1; /* window is allowed to be resized by the window manager */
     Window      embedder;       /* window id of embedder */
     Pixmap         icon_pixmap;
     Pixmap         icon_mask;
@@ -732,7 +741,6 @@ extern XContext cursor_context;
 
 extern BOOL is_current_process_focused(void);
 extern void X11DRV_ActivateWindow( HWND hwnd, HWND previous );
-extern void set_window_cursor( Window window, HCURSOR handle );
 extern void reapply_cursor_clipping(void);
 extern void ungrab_clipping_window(void);
 extern void move_resize_window( HWND hwnd, int dir, POINT pos );
@@ -748,7 +756,7 @@ extern POINT virtual_screen_to_root( INT x, INT y );
 extern POINT root_to_virtual_screen( INT x, INT y );
 extern RECT get_host_primary_monitor_rect(void);
 extern RECT get_work_area( const RECT *monitor_rect );
-extern void xinerama_get_fullscreen_monitors( const RECT *rect, unsigned int *generation, long *indices );
+extern BOOL xinerama_get_fullscreen_monitors( const RECT *rect, unsigned int *generation, long *indices );
 extern void xinerama_init( unsigned int width, unsigned int height );
 extern void init_recursive_mutex( pthread_mutex_t *mutex );
 extern void init_icm_profile(void);
@@ -764,6 +772,21 @@ extern const unsigned int *depths;
 /* Use a distinct type for the settings id, to avoid mixups other types of ids */
 typedef struct { ULONG_PTR id; } x11drv_settings_id;
 
+struct x11drv_mode
+{
+    DEVMODEW    mode;
+    union
+    {
+#ifdef HAVE_X11_EXTENSIONS_XF86VMODE_H
+        XF86VidModeModeInfo mode_info;
+#endif
+#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
+        SizeID              size_id;
+        RRMode              rr_mode;
+#endif
+    };
+};
+
 /* Required functions for changing and enumerating display settings */
 struct x11drv_settings_handler
 {
@@ -788,10 +811,7 @@ struct x11drv_settings_handler
      * dmDisplayFlags and dmDisplayFrequency
      *
      * Return FALSE on failure with parameters unchanged and error code set. Return TRUE on success */
-    BOOL (*get_modes)(x11drv_settings_id id, DWORD flags, DEVMODEW **modes, UINT *mode_count, BOOL full);
-
-    /* free_modes() will be called to free the mode list returned from get_modes() */
-    void (*free_modes)(DEVMODEW *modes);
+    BOOL (*get_modes)(x11drv_settings_id id, DWORD flags, struct x11drv_mode **modes, UINT *mode_count);
 
     /* get_current_mode() will be called to get the current display mode of the device of id
      *
@@ -806,7 +826,7 @@ struct x11drv_settings_handler
      * mode must be a valid mode from get_modes() with optional fields, such as dmPosition set.
      *
      * Return DISP_CHANGE_*, same as ChangeDisplaySettingsExW() return values */
-    LONG (*set_current_mode)(x11drv_settings_id id, const DEVMODEW *mode);
+    LONG (*set_current_mode)(x11drv_settings_id id, const struct x11drv_mode *mode);
 };
 
 #define NEXT_DEVMODEW(mode) ((DEVMODEW *)((char *)((mode) + 1) + (mode)->dmDriverExtra))
@@ -904,13 +924,6 @@ static inline BOOL is_window_rect_mapped( const RECT *rect )
             max( rect->bottom, rect->top + 1 ) > virtual_rect.top);
 }
 
-/* unixlib interface */
-
-extern NTSTATUS x11drv_tablet_attach_queue( void *arg );
-extern NTSTATUS x11drv_tablet_get_packet( void *arg );
-extern NTSTATUS x11drv_tablet_load_info( void *arg );
-extern NTSTATUS x11drv_tablet_info( void *arg );
-
 /* GDI helpers */
 
 static inline BOOL lp_to_dp( HDC hdc, POINT *points, INT count )
